import re
from enum import Enum

from marshmallow import Schema, ValidationError, fields, validate

from aws_iam_generator.utils import ListOrValueField

from .utils import AWSIAMRoleARNField, aws_resource_name_validator

TYPES_REGEX = '(string|integer|list\\(string\\)|list\\(integer\\)|enum\\(\\d+(,\\d+)*\\)|enum\\(\\w+(,\\w+)*\\))'


class VariablesTypes(Enum):
    string = 'string'
    integer = 'integer'
    list_of_strings = 'list(string)'
    list_of_integers = 'list(integer)'
    enum_of_strings = 'enum\\(\\w+(,\\w+)*\\))'
    enum_of_integers = 'enum\\(\\d+(,\\d+)*\\)'

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def get_schema(cls, value):
        _map = {
            'string': fields.String(required=True),
            'integer': fields.Integer(required=True),
            'list(string)': fields.List(fields.String, required=True),
            'list(integer)': fields.List(fields.Integer, required=True)
        }

        if value in cls.list():
            return _map[value]
        elif re.fullmatch(cls.enum_of_integers.value, value):
            values = value.replace('enum(', '').replace(')', '').split(',')
            values = [int(v) for v in values]
            return fields.List(fields.String, validate=validate.OneOf(values))
        elif re.fullmatch(cls.enum_of_strings.value, value):
            values = value.replace('enum(', '').replace(')', '').split(',')
            return fields.List(fields.String, validate=validate.OneOf(values))


class PolicyDocumentStatementValidator(Schema):
    Action = ListOrValueField(fields.String, missing=[])
    Effect = fields.String(required=True)
    Resource = ListOrValueField(fields.String, missing=[])


class RoleValidator(Schema):
    Tags = fields.Mapping(keys=fields.String, values=fields.String, allow_none=True)
    Name = fields.String(required=False, allow_none=True, validate=aws_resource_name_validator)
    Path = fields.String(required=False, allow_none=True, validate=aws_resource_name_validator)
    Trusts = ListOrValueField(fields.String, missing=[])
    ManagedPolicies = ListOrValueField(fields.String, missing=[])
    InAccounts = ListOrValueField(fields.String, missing=[])
    Description = fields.String(allow_none=True)


class ServiceLinkedRoleValidator(Schema):
    ServiceName = fields.String(required=True)
    InAccounts = ListOrValueField(fields.String, missing=[])
    Description = fields.String(allow_none=True)
    Tags = fields.Mapping(keys=fields.String, values=fields.String, allow_none=True)


class PolicyDocumentValidator(Schema):
    Version = fields.String(required=True)
    Statement = ListOrValueField(fields.Nested(PolicyDocumentStatementValidator), missing=[])


class AccountValidator(Schema):
    Description = fields.String(required=False, missing=None)
    AccessRoleArn = AWSIAMRoleARNField(required=False, missing=None)
    AccessRoleName = fields.String(required=False, missing=None)
    Id = fields.String(required=False, missing=None)
    S3BucketKey = fields.String(required=False, missing=None)
    RegionName = fields.String(required=False, missing='us-east-1')


class RegionValidator(Schema):
    Description = fields.String(required=False, missing=None)


class VariableValidator(Schema):
    Description = fields.String(required=False, missing=None)
    Type = fields.String(required=True, validate=validate.Regexp(TYPES_REGEX, error='{input} does not match {regex}'))


class PolicyValidator(Schema):
    Description = fields.String(required=True)
    PolicyDocument = fields.Nested(PolicyDocumentValidator, missing=[])


class AWSIAMRolesSpecificationValidator(Schema):
    Regions = fields.Mapping(keys=fields.String, values=fields.Nested(RegionValidator), default={})
    Accounts = fields.Mapping(keys=fields.String, values=fields.Nested(AccountValidator), default={})
    Variables = fields.Mapping(keys=fields.String, values=fields.Nested(VariableValidator), default={})
    Policies = fields.Mapping(keys=fields.String, values=fields.Nested(PolicyValidator), required=True)
    Roles = fields.Mapping(keys=fields.String(validate=aws_resource_name_validator),
                           values=fields.Nested(RoleValidator), required=True)
    ServiceLinkedRoles = fields.Mapping(keys=fields.String,
                                        values=fields.Nested(ServiceLinkedRoleValidator))


def validate_accounts(data):
    accounts = set(data['Accounts'].keys())  # mention deploy accounts here
    deploy_accounts = set()  # specify access for this

    for k, v in data.get('Roles', {}).items():
        deploy_accounts.update(set([i.split('.')[1] for i in v['InAccounts']]))

    for k, v in data.get('ServiceLinkedRoles', {}).items():
        deploy_accounts.update(set([i.split('.')[1] for i in v['InAccounts']]))

    if not accounts.intersection(deploy_accounts) == deploy_accounts:
        raise ValidationError(message={"Roles": [
            f'Following accounts were mentioned in Roles but not in Accounts '
            f'{", ".join(deploy_accounts.difference(accounts))}']})

    return data



import re
import typing

from marshmallow import Schema, ValidationError, fields, validate

VALID_AWS_RESOURCE_NAMES = re.compile(r'^[a-zA-Z0-9]+$')

aws_resource_name_validator = validate.Regexp(VALID_AWS_RESOURCE_NAMES, error='{input} does not match {regex}')


def parse_account_id(account_id):
    if len(str(account_id)) < 12:
        return ('0' * (len(str(account_id)) - 12)) + str(account_id)
    return account_id


class ListOrValueField(fields.List):
    def __init__(self, cls_or_instance: typing.Union[fields.Field, type], **kwargs) -> None:
        super().__init__(cls_or_instance, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> typing.Optional[typing.List[typing.Any]]:
        if isinstance(value, str):
            return super()._serialize([value], attr, obj, **kwargs)
        return super()._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs) -> typing.List[typing.Any]:
        if isinstance(value, str):
            return super()._deserialize([value], attr, data, **kwargs)
        return super()._deserialize(value, attr, data, **kwargs)

    def _validated(self, value):
        if isinstance(value, str):
            return self._validate([value])
        else:
            return self._validate(value)


class IntOrString(fields.String):

    def _serialize(self, value, attr, obj, **kwargs) -> typing.Optional[typing.List[typing.Any]]:
        if isinstance(value, int):
            return super()._serialize(str(value), attr, obj, **kwargs)
        return super()._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs) -> typing.List[typing.Any]:
        if isinstance(value, int):
            return super()._deserialize(str(value), attr, data, **kwargs)
        return super()._deserialize(value, attr, data, **kwargs)

    def _validated(self, value):
        if isinstance(value, int):
            return self._validate(str(value))
        else:
            return self._validate(value)


class OneOfSchema(Schema):
    type_field = "type"
    type_field_remove = True
    type_schemas = {}

    def get_obj_type(self, obj):
        """Returns name of object schema"""
        return obj.__class__.__name__

    def dump(self, obj, *, many=None, **kwargs):
        errors = {}
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if not many:
            result = result_data = self._dump(obj, **kwargs)
        else:
            for idx, o in enumerate(obj):
                try:
                    result = self._dump(o, **kwargs)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.normalized_messages()
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=obj, valid_data=result)
            raise exc

    def _dump(self, obj, *, update_fields=True, **kwargs):
        obj_type = self.get_obj_type(obj)
        if not obj_type:
            return (
                None,
                {"_schema": "Unknown object class: %s" % obj.__class__.__name__},
            )

        type_schema = self.type_schemas.get(obj_type)
        if not type_schema:
            return None, {"_schema": "Unsupported object type: %s" % obj_type}

        schema = type_schema if isinstance(type_schema, Schema) else type_schema()

        schema.context.update(getattr(self, "context", {}))

        result = schema.dump(obj, many=False, **kwargs)
        if result is not None:
            result[self.type_field] = obj_type
        return result

    def load(self, data, *, many=None, partial=None, unknown=None, **kwargs):
        errors = {}
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if partial is None:
            partial = self.partial
        if not many:
            try:
                result = result_data = self._load(
                    data, partial=partial, unknown=unknown, **kwargs
                )
                #  result_data.append(result)
            except ValidationError as error:
                result_errors = error.normalized_messages()
                result_data.append(error.valid_data)
        else:
            for idx, item in enumerate(data):
                try:
                    result = self._load(item, partial=partial, **kwargs)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.normalized_messages()
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=data, valid_data=result)
            raise exc

    def _load(self, data, *, partial=None, unknown=None, **kwargs):
        if not isinstance(data, dict):
            raise ValidationError({"_schema": "Invalid data type: %s" % data})

        data = dict(data)
        unknown = unknown or self.unknown

        data_type = data.get(self.type_field)
        if self.type_field in data and self.type_field_remove:
            data.pop(self.type_field)

        if not data_type:
            raise ValidationError(
                {self.type_field: ["Missing data for required field."]}
            )

        try:
            type_schema = self.type_schemas.get(data_type)
        except TypeError:
            # data_type could be unhashable
            raise ValidationError({self.type_field: ["Invalid value: %s" % data_type]})
        if not type_schema:
            raise ValidationError(
                {self.type_field: ["Unsupported value: %s" % data_type]}
            )

        schema = type_schema if isinstance(type_schema, Schema) else type_schema()

        schema.context.update(getattr(self, "context", {}))

        return schema.load(data, many=False, partial=partial, unknown=unknown, **kwargs)

    def validate(self, data, *, many=None, partial=None):
        try:
            self.load(data, many=many, partial=partial)
        except ValidationError as ve:
            return ve.messages
        return {}


class AWSIAMRoleARNField(fields.String):
    def __init__(self, *args, **kwargs):
        pattern = 'arn:aws:iam::\\d{12}:role/.+'
        arn_validator = validate.Regexp(pattern, error='{input} does not match ARN format {regex}')
        super().__init__(*args, validate=arn_validator, example='arn:aws:iam::123456789012:role/ExampleRoleName',
                         description='AWS IAM Role ARN', **kwargs)


def patch_dict(a: dict, b: dict):
    for k1, v1 in b.items():
        for k2, v2 in v1.asdict().items():
            if k1 not in a:
                a[k1] = {}
            if v2 and k2 not in ("Description", "AccessRoleArn", "Type"):
                a[k1][k2] = v2



import typing
from dataclasses import dataclass, field
from os import path

from marshmallow import post_load

from .validators import RoleValidator, ServiceLinkedRoleValidator


class RoleException(Exception):
    pass


@dataclass
class Role:
    Trusts: typing.List[typing.AnyStr]
    Name: typing.AnyStr
    ManagedPolicies: typing.List[typing.AnyStr]
    InAccounts: typing.List[typing.AnyStr]
    Tags: typing.Dict[typing.AnyStr, typing.AnyStr] = field(default_factory=dict)
    _InAccounts: typing.List[typing.AnyStr] = field(default_factory=list)
    Description: typing.AnyStr = field(default=None)
    Path: typing.AnyStr = field(default='')

    @property
    def Arn(self):
        return [f'arn:aws:iam::{i}:role/{path.join(self.Path, self.Name)}' for i in self._InAccounts]


@dataclass
class ServiceLinkedRole:
    ServiceName: typing.AnyStr
    InAccounts: typing.List[typing.AnyStr]
    Description: typing.AnyStr = field(default=None)
    Tags: typing.Dict[typing.AnyStr, typing.AnyStr] = field(default_factory=dict)


class RoleSerializer(RoleValidator):

    @post_load
    def load2obj(self, data, *args, **kwargs):
        return Role(**data)


class ServiceLinkedRoleSerializer(ServiceLinkedRoleValidator):
    @post_load
    def load2obj(self, data, *args, **kwargs):
        return ServiceLinkedRole(**data)



import json
import os

import pytest

from aws_iam_generator.serializers import AWSIAMRolesSpecificationSerializer


@pytest.fixture(scope='session')
def example_2_roles_specification():
    BASE_DIR = os.path.dirname(os.path.realpath(__file__))
    return json.load(open(os.path.join(BASE_DIR, 'input.json')))


@pytest.fixture(scope='session')
def serialized_example_2_roles_specification(example_2_roles_specification):
    obj = AWSIAMRolesSpecificationSerializer().load(data=example_2_roles_specification)
    obj.set_parameters(
        Accounts={
            'AppsAccount': {
                "Id": '123456789012',
                "AccessRoleName": "DefaultAccessRole"
            },
            'PipelineAccount': {
                "Id": '456789012345',
                "AccessRoleName": "DefaultAccessRole"
            },
            'WorkloadUserAccount': {
                "Id": '789012345678',
                "AccessRoleName": "DefaultAccessRole"
            }
        },
        Regions={}
    )
    return obj


@pytest.fixture(scope='session')
def templates(serialized_example_2_roles_specification):
    serialized_example_2_roles_specification.Reference = ""
    return serialized_example_2_roles_specification.generate_cloudformation_templates()


@pytest.fixture(scope='session')
def workload_user_account_cf_template(templates):
    for template in templates:
        if template['Name'] == 'WorkloadUserAccount-':
            return template


@pytest.fixture(scope='session')
def apps_account_slr_cf_template(templates):
    for template in templates:
        if template['Name'] == 'AppsAccount--Service-Linked-Roles':
            return template



import json
import os


def test_valid_workload_user_account_cf_template(workload_user_account_cf_template):
    BASE_DIR = os.path.dirname(os.path.realpath(__file__))
    expected = json.load(open(os.path.join(BASE_DIR, 'WorkloadUserAccount.json')))
    assert workload_user_account_cf_template['Body'] == expected


def test_valid_apps_account_slr_cf_template(apps_account_slr_cf_template):
    BASE_DIR = os.path.dirname(os.path.realpath(__file__))
    expected = json.load(open(os.path.join(BASE_DIR, 'AppsAccount-Service-Linked-Roles.json')))
    assert apps_account_slr_cf_template['Body'] == expected



from moto import mock_cloudformation, mock_iam, mock_sts

from aws_iam_generator import AWSIAMGenerator


@mock_iam
@mock_cloudformation
@mock_sts
def test_serializer_roles_spec(example_roles_specification):
    generator = AWSIAMGenerator(reference_name='test')
    generator.load_spec(spec=example_roles_specification)
    generator.set_parameters(
        Accounts={
            'AppsAccount': {
                "Id": '112345678901',
                "AccessRoleName": "DefaultName"
            },
            'PipelineAccount': {
                "Id": '212345678901'
            }
        },
        Regions={
            'PipelineRegion': {
                "Id": "us-east-1"
            },
            'AppsRegion': {
                "Id": "us-east-1"
            }
        },
        Variables={
            "HostedZoneID": {
                "Value": "*"
            },
            "TrustRolesArns": {
                "Value": ["arn:aws:iam::123456789011:role/ExternalAccessRole"]
            }
        }
    )



import re
import typing
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from uuid import uuid4

from marshmallow import Schema, fields, missing, post_load, pre_load, validate
from marshmallow.exceptions import ValidationError
from troposphere import Export, GetAtt, Ref, Sub, Template

from .account import Account, AccountSerializer, Region, RegionSerializer, Variable, VariableSerializer
from .aws import ManagedPolicy, Output
from .aws import Role as AWSRole
from .aws import ServiceLinkedRole as AWSServiceLinkedRole
from .policy import Policy, PolicySerializer
from .role import Role, RoleSerializer, ServiceLinkedRole, ServiceLinkedRoleSerializer
from .utils import AWSIAMRoleARNField, aws_resource_name_validator, parse_account_id, patch_dict
from .validators import AWSIAMRolesSpecificationValidator, VariablesTypes, validate_accounts

REF_PATTERN = "(.*)"
VAR_PATTERN = "(?<={{)([^x}}]+)(?=}})"
SUB_PATTERN = "(?<={)([^x}]+)(?=})"


class IAMRolesTypes(Enum):
    GENERIC = 'GENERIC'
    SERVICE_LINKED_ROLES = 'SERVICE_LINKED_ROLES'


class AccountIdSerializer(Schema):
    Id = fields.String(required=True, validate=validate.Regexp(r'\d{12}'))
    AccessRoleArn = AWSIAMRoleARNField(required=False, missing=None)
    S3BucketKey = fields.String(required=False, missing=None)
    RegionName = fields.String(required=False, missing='us-east-1')
    AccessRoleName = fields.String(required=False, missing=None)

    @post_load
    def load2obj(self, data, *args, **kwargs):
        if not self.context.get('RequireAccessRole', False):
            pass
        else:
            if not (data['AccessRoleArn'] or data['AccessRoleName']):
                raise ValidationError(message={
                    "Error": ["Account was not provided with AccessRoleName or AccessRoleArn."]
                })
            elif data.get('AccessRoleName'):
                data['AccessRoleArn'] = f'arn:aws:iam::{data["Id"]}:role/' \
                                        f'{data.pop("AccessRoleName")}'

        data.pop('AccessRoleName', None)
        data['Id'] = parse_account_id(data['Id'])
        return Account(**data)


class VariableValueSerializer(Schema):
    @classmethod
    def get_value_serializer_class(cls, variable: Variable):
        class GeneratedSchema(Schema):
            Value = VariablesTypes.get_schema(variable.Type)

            @post_load
            def load2obj(self, data, *args, **kwargs):
                return Variable(**data)

        return GeneratedSchema


class RegionIdSerializer(Schema):
    Id = fields.String(required=True)

    @post_load
    def load2obj(self, data, *args, **kwargs):
        return Region(**data)


@dataclass(frozen=False)
class AWSIAMRolesSpecification:
    Accounts: typing.Dict[str, Account]
    Policies: typing.Dict[str, Policy]
    Roles: typing.Dict[str, Role]
    Variables: typing.Dict[str, Variable] = field(default_factory=dict)
    Regions: typing.Dict[str, Region] = field(default_factory=dict)
    ServiceLinkedRoles: typing.Dict[str, ServiceLinkedRole] = field(default_factory=dict)
    Reference: typing.AnyStr = field(default=uuid4().hex)

    @property
    def build_version(self):
        return datetime.utcnow().strftime("%Y-%m-%dZ%H:%M:%S")

    def parse(self):
        for key, policy in self.Policies.items():
            for statement in policy.PolicyDocument.Statement:
                for i in range(len(statement.Resource)):
                    _vars = re.compile(VAR_PATTERN).findall(statement.Resource[i])
                    sub = statement.Resource[i]
                    for var in _vars:
                        _id = self.get_variable(var)
                        sub = sub.replace("{{" + var + "}}", str(_id))
                    statement.Resource[i] = Sub(sub) if re.compile(SUB_PATTERN).findall(statement.Resource[i]) else sub

        for role in self.Roles.values():
            role._InAccounts = [self.get_variable(i) for i in role.InAccounts]

        for role in self.Roles.values():
            parsed_trusts = []

            for i in range(len(role.Trusts)):
                if 'Accounts.' in role.Trusts[i]:
                    parsed_trusts.append(self.get_variable(role.Trusts[i]))
                elif 'Roles.' in role.Trusts[i]:
                    parsed_trusts += self.get_variable(role.Trusts[i])
                elif 'Variables.' in role.Trusts[i]:
                    value = self.get_variable(role.Trusts[i])
                    if type(value) == list:
                        parsed_trusts += value
                    else:
                        parsed_trusts.append(value)
                elif re.compile(VAR_PATTERN).findall(role.Trusts[i]):
                    _id = self.get_variable(role.Trusts[i])
                    if type(_id) == list:
                        parsed_trusts += _id
                    else:
                        parsed_trusts.append(_id)
                else:
                    parsed_trusts.append(role.Trusts[i])

            role.Trusts = parsed_trusts

            for i in range(len(role.ManagedPolicies)):
                if '.' in role.ManagedPolicies[i]:
                    ref = role.ManagedPolicies[i].split('.')[-1]
                    role.ManagedPolicies[i] = Ref(ref)

    def get_variable(self, path):
        try:
            _path = path.replace('{{', '').replace('}}', '').split('.')
            value = getattr(getattr(self, _path[0])[_path[1]], _path[2])
        except Exception as e:
            raise ValidationError(message=f'Could not find variable with path: {path}') from e
        return value

    def set_parameters(self, Accounts={}, Regions={}, Variables={}):
        accounts_fields_map = {
            k: fields.Nested(
                AccountIdSerializer(
                    context={'RequireAccessRole': v.RequireAccessRole, "Name": k}
                ),
                required=False if v.Id else True,
                missing=v if v.Id else missing,
            ) for k, v
            in self.Accounts.items()
        }
        regions_fields_map = {
            k: fields.Nested(
                RegionIdSerializer,
                required=False if v.Id else True,
                missing=v if v.Id else missing,
            ) for k, v in self.Regions.items()
        }

        variables_fields_map = {
            k: fields.Nested(
                VariableValueSerializer.get_value_serializer_class(self.Variables[k]),
                required=True
            ) for k in self.Variables.keys()
        }

        accounts_schema = Schema.from_dict(accounts_fields_map)
        regions_schema = Schema.from_dict(regions_fields_map)
        variables_schema = Schema.from_dict(variables_fields_map)

        validator = Schema.from_dict({
            'Accounts': fields.Nested(accounts_schema),
            'Regions': fields.Nested(regions_schema),
            'Variables': fields.Nested(variables_schema)
        })
        patch_dict(Accounts, self.Accounts)
        patch_dict(Regions, self.Regions)
        patch_dict(Variables, self.Variables)

        result = validator().load(data={
            'Accounts': Accounts,
            'Regions': Regions,
            'Variables': Variables
        })
        self.Accounts = result['Accounts']
        self.Regions = result['Regions']
        self.Variables = result['Variables']
        self.parse()
        return self.generate_cloudformation_templates()

    def generate_cloudformation_templates(self):
        templates = {k: Template() for k, v in self.Accounts.items()}
        slr_templates = {k: Template() for k, v in self.Accounts.items()}
        templates_names = {}
        slr_templates_names = {}

        for k, v in self.Accounts.items():
            templates_names[k] = v.asdict()
            templates_names[k].update({
                "Name": f'{k}-{self.Reference}'
            })

        for k, v in self.Accounts.items():
            slr_templates_names[k] = v.asdict()
            slr_templates_names[k].update({
                "Name": f'{k}-{self.Reference}-Service-Linked-Roles'
            })

        for role_name, role in self.Roles.items():
            service = []
            aws = []
            for i in role.Trusts:
                if isinstance(i, str) and "amazonaws" in i:
                    if i not in service:
                        service.append(i)
                elif isinstance(i, list):
                    for j in i:
                        if isinstance(j, str) and "amazonaws" in i:
                            if i not in service:
                                service.append(j)
                        else:
                            if re.match("\\d{12}", j) or j == "*":
                                aws.append(f"arn:aws:iam::{j}:root")
                            else:
                                aws.append(j)
                else:
                    if isinstance(i, str) and "amazonaws" in i:
                        service.append(i)
                    else:
                        try:
                            if re.match("\\d{12}", i) or i == "*":
                                aws.append(f"arn:aws:iam::{i}:root")
                            else:
                                aws.append(i)
                        except TypeError:
                            raise ValidationError(f'Value {i} does not match AWSAccountId or IAM Role Arn format.')

            assume_role_policy_document = {
                "Statement": [
                    {
                        "Action": "sts:AssumeRole",
                        "Effect": "Allow",
                        "Principal": {}
                    }
                ],
                "Version": "2012-10-17"
            }
            if service:
                assume_role_policy_document["Statement"][0]["Principal"]['Service'] = service
            if aws:
                assume_role_policy_document["Statement"][0]["Principal"]['AWS'] = aws
            for managed_policy_ref in role.ManagedPolicies:
                if isinstance(managed_policy_ref, Ref):
                    managed_policy_name = managed_policy_ref.to_dict()['Ref']
                    try:
                        policy = self.Policies[managed_policy_name]
                    except KeyError:
                        raise ValidationError(
                            {
                                'Roles': [f'Cloud not find policy template: {managed_policy_name}']
                            }
                        )
                    for i in role.InAccounts:
                        account_name = i.split('.')[1]
                        if managed_policy_name not in templates[account_name].resources:
                            templates[account_name].add_resource(ManagedPolicy(
                                managed_policy_name,
                                Description=policy.Description,
                                Groups=[],
                                ManagedPolicyName=managed_policy_name,
                                PolicyDocument=policy.PolicyDocument.asdict()
                            ))

            for i in role.InAccounts:
                account_name = i.split('.')[1]
                name = role_name
                templates[account_name].add_resource(AWSRole(
                    role_name,
                    AssumeRolePolicyDocument=assume_role_policy_document,
                    ManagedPolicyArns=role.ManagedPolicies,
                    RoleName=name
                ))
                templates[account_name].add_output([
                    Output(
                        role_name + "Arn",
                        Description=role.Description or '',
                        Value=GetAtt(role_name, 'Arn'),
                        Export=Export(Sub("${AWS::StackName}-" + role_name + "Arn"))
                    )
                ])

        for role_name, role in self.ServiceLinkedRoles.items():
            for i in role.InAccounts:
                account_name = i.split('.')[1]
                role_name = role_name.replace('_', '')
                slr_templates[account_name].add_resource(AWSServiceLinkedRole(
                    role_name,
                    Description=role.Description or '',
                    AWSServiceName=role.ServiceName
                ))
                slr_templates[account_name].add_output([
                    Output(
                        role_name + "Ref",
                        Description=role.Description or '',
                        Value=Ref(role_name),
                        Export=Export(Sub("${AWS::StackName}-" + role_name + "Ref"))
                    )
                ])

        result = []
        for account_name, template in templates.items():
            if template.resources:
                result.append(
                    {
                        'AWSAccountID': templates_names[account_name]["Id"],
                        'AccessRoleArn': templates_names[account_name]["AccessRoleArn"],
                        'RegionName': templates_names[account_name]["RegionName"],
                        'S3BucketKey': templates_names[account_name]["S3BucketKey"],
                        'Name': templates_names[account_name]["Name"],
                        'Body': template.to_dict(),
                        "Type": IAMRolesTypes.GENERIC.value
                    }
                )

        for account_name, template in slr_templates.items():
            if template.resources:
                result.append(
                    {
                        'AWSAccountID': slr_templates_names[account_name]["Id"],
                        'AccessRoleArn': slr_templates_names[account_name]["AccessRoleArn"],
                        'RegionName': slr_templates_names[account_name]["RegionName"],
                        'S3BucketKey': slr_templates_names[account_name]["S3BucketKey"],
                        'Name': slr_templates_names[account_name]["Name"],
                        'Body': template.to_dict(),
                        "Type": IAMRolesTypes.SERVICE_LINKED_ROLES.value
                    }
                )

        return result


class AWSIAMRolesSpecificationSerializer(AWSIAMRolesSpecificationValidator):
    Accounts = fields.Mapping(keys=fields.String, values=fields.Nested(AccountSerializer), default={})
    Regions = fields.Mapping(keys=fields.String, values=fields.Nested(RegionSerializer), default={})
    Variables = fields.Mapping(keys=fields.String, values=fields.Nested(VariableSerializer), default={})
    Policies = fields.Mapping(keys=fields.String, values=fields.Nested(PolicySerializer))
    Roles = fields.Mapping(keys=fields.String(validate=aws_resource_name_validator),
                           values=fields.Nested(RoleSerializer))
    ServiceLinkedRoles = fields.Mapping(keys=fields.String,
                                        values=fields.Nested(ServiceLinkedRoleSerializer), default={})

    @pre_load
    def pre_load_validate(self, data, *args, **kwargs):
        validate_accounts(data)

        for k, v in data['Roles'].items():
            if 'Name' not in v:
                v['Name'] = k

        return data

    @post_load
    def load2obj(self, data, *args, **kwargs):
        accounts = set(data['Accounts'].keys())
        deploy_accounts = set()

        for k, v in data.get('Roles', {}).items():
            deploy_accounts.update(set([i.split('.')[1] for i in v.InAccounts]))
        for k, v in data.get('Roles', {}).items():
            deploy_accounts.update(set([i.split('.')[1] for i in v.InAccounts]))
        for k, v in data.get('ServiceLinkedRoles', {}).items():
            deploy_accounts.update(set([i.split('.')[1] for i in v.InAccounts]))

        for acc in accounts.difference(deploy_accounts):
            data['Accounts'][acc].RequireAccessRole = False

        return AWSIAMRolesSpecification(**data)



import json
import os

import pytest

from aws_iam_generator.serializers import AWSIAMRolesSpecificationSerializer


@pytest.fixture(scope='session')
def example_4_roles_specification():
    BASE_DIR = os.path.dirname(os.path.realpath(__file__))
    return json.load(open(os.path.join(BASE_DIR, 'input.json')))


@pytest.fixture(scope='session')
def serialized_example_4_roles_specification(example_4_roles_specification):
    obj = AWSIAMRolesSpecificationSerializer().load(data=example_4_roles_specification)
    obj.set_parameters(
        Accounts={
            "SharedServicesAccount": {
                "Id": "098765432109",
                "AccessRoleName": "DefaultAccessRole"
            }
        },
        Regions={},
        Variables={
            'HostedZoneID': {
                "Value": "1234567890"
            },
            'TrustAccountsIds': {
                "Value": ["123456789012"]
            },
            'TrustRolesArns': {
                "Value": ["arn:aws:iam::123456789012:role/ExampleRole"]
            }
        }
    )
    return obj


@pytest.fixture(scope='session')
def templates(serialized_example_4_roles_specification):
    serialized_example_4_roles_specification.Reference = ""

    return serialized_example_4_roles_specification.generate_cloudformation_templates()


@pytest.fixture(scope='session')
def shared_services_account_cf_template(templates):
    for template in templates:
        if template['Name'] == 'SharedServicesAccount-':
            return template


@pytest.fixture(scope='session')
def example_4_shared_services_account_template():
    BASE_DIR = os.path.dirname(os.path.realpath(__file__))
    return json.load(open(os.path.join(BASE_DIR, 'SharedServicesAccount.json')))



from troposphere import Output as AWSOutput
from troposphere.iam import ManagedPolicy as AWSManagedPolicy
from troposphere.iam import Role as AWSRole
from troposphere.iam import ServiceLinkedRole as AWSServiceLinkedRole

from .utils import VALID_AWS_RESOURCE_NAMES


class Output(AWSOutput):
    def validate_title(self):
        if not VALID_AWS_RESOURCE_NAMES.match(self.title):
            raise ValueError(f'Name {self.title} not not match {VALID_AWS_RESOURCE_NAMES}')


class ManagedPolicy(AWSManagedPolicy):
    def validate_title(self):
        if not VALID_AWS_RESOURCE_NAMES.match(self.title):
            raise ValueError(f'Name {self.title} not not match {VALID_AWS_RESOURCE_NAMES}')


class Role(AWSRole):
    def validate_title(self):
        if not VALID_AWS_RESOURCE_NAMES.match(self.title):
            raise ValueError(f'Name {self.title} not not match {VALID_AWS_RESOURCE_NAMES}')


class ServiceLinkedRole(AWSServiceLinkedRole):
    def validate_title(self):
        if not VALID_AWS_RESOURCE_NAMES.match(self.title):
            raise ValueError(f'Name {self.title} not not match {VALID_AWS_RESOURCE_NAMES}')



import typing
from dataclasses import dataclass

from marshmallow import fields, post_load

from aws_iam_generator.utils import ListOrValueField

from .validators import PolicyDocumentStatementValidator, PolicyDocumentValidator, PolicyValidator


@dataclass
class PolicyDocumentStatement:
    Action: typing.List
    Effect: typing.AnyStr
    Resource: typing.List

    def asdict(self):
        return {
            "Action": self.Action,
            "Effect": self.Effect,
            "Resource": self.Resource
        }


class PolicyDocumentStatementSerializer(PolicyDocumentStatementValidator):

    @post_load
    def load2obj(self, data, *args, **kwargs):
        return PolicyDocumentStatement(**data)


@dataclass
class PolicyDocument:
    Version: typing.AnyStr
    Statement: typing.List[PolicyDocumentStatement]

    def asdict(self):
        return {
            "Version": self.Version,
            "Statement": [i.asdict() for i in self.Statement]
        }


class PolicyDocumentSerializer(PolicyDocumentValidator):
    Statement = ListOrValueField(fields.Nested(PolicyDocumentStatementSerializer), missing=[])

    @post_load
    def load2obj(self, data, *args, **kwargs):
        return PolicyDocument(**data)


@dataclass
class Policy:
    Description: typing.AnyStr
    PolicyDocument: PolicyDocument


class PolicySerializer(PolicyValidator):
    PolicyDocument = fields.Nested(PolicyDocumentSerializer, missing=[])

    @post_load
    def load2obj(self, data, *args, **kwargs):
        return Policy(**data)



import typing
from dataclasses import dataclass, field

from marshmallow import post_load

from .validators import AccountValidator, RegionValidator, VariableValidator


@dataclass
class Account:
    Description: typing.AnyStr = field(default=None)
    Id: typing.AnyStr = field(default=None)
    AccessRoleArn: typing.AnyStr = field(default=None)
    RegionName: typing.AnyStr = field(default='us-east-1')
    S3BucketKey: typing.AnyStr = field(default=None)
    RequireAccessRole: bool = field(default=True)

    def asdict(self):
        return {
            "Id": self.Id,
            "Description": self.Description,
            "S3BucketKey": self.S3BucketKey,
            "RegionName": self.RegionName,
            "AccessRoleArn": self.AccessRoleArn,
            "AccessRoleName": self.AccessRoleArn.rsplit('/')[-1] if self.AccessRoleArn else None
        }


@dataclass
class Region:
    Description: typing.AnyStr = field(default=None)
    Id: typing.AnyStr = field(default=None)

    def asdict(self):
        return {
            "Id": self.Id,
            "Description": self.Description,
        }


@dataclass
class Variable:
    Description: typing.AnyStr = field(default=None)
    Value: typing.AnyStr = field(default=None)
    Type: typing.AnyStr = field(default=None)

    def asdict(self):
        return {
            "Value": self.Value,
            "Description": self.Description,
            "Type": self.Type
        }


class AccountSerializer(AccountValidator):

    @post_load
    def load2obj(self, data, *args, **kwargs):
        if data["AccessRoleName"]:
            data['AccessRoleArn'] = f'arn:aws:iam::{data["Id"]}:role/' \
                                    f'{data["AccessRoleName"]}'
        data.pop("AccessRoleName")
        return Account(**data)


class RegionSerializer(RegionValidator):

    @post_load
    def load2obj(self, data, *args, **kwargs):
        return Region(**data)


class VariableSerializer(VariableValidator):

    @post_load
    def load2obj(self, data, *args, **kwargs):
        return Variable(**data)



import codecs
import os.path

from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    return codecs.open(os.path.join(HERE, *parts), 'r').read()


setup(
    name='aws-iam-generator',
    version='0.9.3',
    packages=find_packages(exclude=(
        'tests'
    )),
    install_requires=[
        "marshmallow==3.3.0",
        "troposphere==2.6.3 ",
        "PyYAML==5.3"
    ],
    author="Michal Murawski",
    author_email="mmurawski777@gmail.com",
    description="Utilities package for pynamodb.",
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    url="https://github.com/micmurawski/aws-iam-generator",
    include_package_data=True,
    python_requires='>=3.6',
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
)



import json
import os


def test_valid_shared_services_account_cf_template(shared_services_account_cf_template):
    BASE_DIR = os.path.dirname(os.path.realpath(__file__))
    expected = json.load(open(os.path.join(BASE_DIR, 'SharedServicesAccount.json')))
    assert shared_services_account_cf_template['Body'] == expected



import json
from datetime import datetime
from uuid import uuid4

import boto3

from .serializers import AWSIAMRolesSpecificationSerializer


class AWSIAMGeneratorException(Exception):
    pass


class AWSIAMGenerator:

    def __init__(self, session=None, role_session_name=None, reference_name=None, unique_id=False):
        self.spec = None
        self.output = None
        self.unique_id = unique_id
        self.reference_name = uuid4().hex if reference_name is None else reference_name
        self.role_session_name = f'aws-iam-generator-{datetime.utcnow().strftime("%Y%m%d%H%M%S")}' \
            if role_session_name is None else role_session_name
        self.session = session if session else boto3.Session()

    def load_spec(self, spec):
        self.spec = AWSIAMRolesSpecificationSerializer().load(data=spec)
        self.spec.UniqueID = self.unique_id
        self.spec.Reference = self.reference_name

    def set_parameters(self, **kwargs):
        self.output = self.spec.set_parameters(**kwargs)

    def deploy(self):

        if self.output is None:
            raise AWSIAMGeneratorException('Specification wasn\'t loaded yet. Please execute load_spec method before.')

        result = {}
        for obj in self.output:
            sts_client = self.session.client('sts')
            sts_response = sts_client.assume_role(
                RoleArn=obj['AccessRoleArn'],
                RoleSessionName=self.role_session_name
            )
            cf_client = self.session.client(
                'cloudformation',
                aws_access_key_id=sts_response["Credentials"]["AccessKeyId"],
                aws_secret_access_key=sts_response["Credentials"]["SecretAccessKey"],
                aws_session_token=sts_response["Credentials"]["SessionToken"],
                region_name=obj['RegionName']
            )

            cf_request_body = {
                "StackName": obj['Name'],
                "Parameters": (),
                "OnFailure": 'DO_NOTHING',
                "TimeoutInMinutes": 2,
                "Capabilities": [
                    'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM', 'CAPABILITY_AUTO_EXPAND',
                ]
            }

            if obj['S3BucketKey']:
                bucket, key = obj['S3BucketKey'].replace('s3://', '').split('/', 1)
                s3_client = self.session.client('s3')
                s3_client.put_object(
                    Bucket=bucket,
                    Key=f'{key}/{obj["Name"]}.json',
                    Body=json.dumps(obj['Body'])
                )
                cf_request_body['TemplateURL'] = f'https://{bucket}.s3.amazonaws.com/{key}'
            else:
                cf_request_body['TemplateBody'] = json.dumps(obj['Body'])

            cf_response = cf_client.create_stack(
                **cf_request_body
            )
            result[obj['Name']] = cf_response

            return result
